use std::{
    env::current_dir, fs::{self, create_dir_all, exists}, io::Write, path::{Path, PathBuf}
};

use git2::Repository;
use log::{debug, info, trace, warn};
use serde::{Deserialize, Serialize};
use sha2::{Digest, Sha256, Sha512};

use crate::prelude::abspath;

#[derive(Serialize, Deserialize, Debug)]
pub struct Git {
    pub url: String,
    #[serde(default = "default_branch")]
    pub branch: String,
}

fn default_branch() -> String {
    "main".to_string()
}

impl Git {
    fn url_hash(&self) -> String {
        let mut hasher = Sha256::new();
        hasher.update(&self.url);
        let hash = hasher.finalize();
        hex::encode(hash)
    }

    fn branch_hash(&self) -> String {
        let mut hasher = Sha256::new();
        hasher.update(&self.url);
        let hash = hasher.finalize();
        hex::encode(hash)
    }

    pub fn repository_path_str(&self) -> String {
        format!(".data/git/{}{}", self.url_hash(), self.branch_hash())
    }

    pub fn repository_path(&self) -> Result<PathBuf, std::io::Error> {
        fs::canonicalize(self.repository_path_str())
    }

    pub fn exists_on_fs(&self) -> bool {
        self.repository_path().is_ok()
    }

    pub fn clone_repository(&self) -> Result<Repository, git2::Error> {
        Repository::clone_recurse(&self.url, Path::new(&self.repository_path_str()))
    }

    pub fn repository(&self) -> Result<Repository, git2::Error> {
        if !self.exists_on_fs() {
            create_dir_all(self.repository_path_str()).unwrap();
        }
        match Repository::open(self.repository_path().unwrap()) {
            Ok(r) => Ok(r),
            Err(_) => self.clone_repository(),
        }
    }

    fn up_to_date(&self) -> Result<bool, Box<dyn std::error::Error>>{
        debug!("checking repo up to date...");
        let repo = self.repository()?;
        let mut remote = repo.find_remote("origin")?;

        // Fetch latest references from remote
        remote.fetch(&[self.branch.clone()], None, None)?;

        let fetch_head = repo.refname_to_id(&format!("refs/remotes/origin/{}", self.branch))?;
        let local_head = repo.refname_to_id(&format!("refs/heads/{}", self.branch))?;


        Ok(fetch_head == local_head)
    }

    pub fn update(&self) -> Result<bool, Box<dyn std::error::Error>> {
        if self.up_to_date()? {
            return Ok(true);
        }
        debug!("updating repository...");
        let repository = self.repository()?;
        let mut remote = repository.find_remote("origin")?;
        let mut cb = git2::RemoteCallbacks::new();

        cb.transfer_progress(|stats| {
            if stats.received_objects() == stats.total_objects() {
                print!(
                    "resolving deltas {}/{}\r",
                    stats.indexed_deltas(),
                    stats.total_deltas()
                );
            } else if stats.total_objects() > 0 {
                print!(
                    "received {}/{} objects ({}) in {} bytes\r",
                    stats.received_objects(),
                    stats.total_objects(),
                    stats.indexed_objects(),
                    stats.received_bytes()
                );
            }
            std::io::stdout().flush().unwrap();
            true
        });

        let mut fo = git2::FetchOptions::new();
        fo.remote_callbacks(cb);
        // Always fetch all tags.
        // Perform a download and also update tips
        fo.download_tags(git2::AutotagOption::All);
        info!("fetching {}...", remote.name().unwrap());
        remote.fetch(&[self.branch.clone()], Some(&mut fo), None)?;

        let fetch_head = repository.find_reference("FETCH_HEAD")?;

        do_merge(
            &repository,
            &self.branch,
            repository.reference_to_annotated_commit(&fetch_head)?,
        )?;

        Ok(true)
    }

    pub fn ensure(&self) -> Result<&Self, Box<dyn std::error::Error>> {
        if self.exists_on_fs() {
            self.update();
        } else {
            self.clone_repository()?;
        }
        Ok(self)
    }
}

fn do_merge<'a>(
    repo: &'a Repository,
    remote_branch: &str,
    fetch_commit: git2::AnnotatedCommit<'a>,
) -> Result<(), git2::Error> {
    // 1. do a merge analysis
    let analysis = repo.merge_analysis(&[&fetch_commit])?;

    // 2. Do the appopriate merge
    if analysis.0.is_fast_forward() {
        info!("doing a fast forward...");
        // do a fast forward
        let refname = format!("refs/heads/{}", remote_branch);
        match repo.find_reference(&refname) {
            Ok(mut r) => {
                fast_forward(repo, &mut r, &fetch_commit)?;
            }
            Err(_) => {
                // The branch doesn't exist so just set the reference to the
                // commit directly. Usually this is because you are pulling
                // into an empty repository.
                repo.reference(
                    &refname,
                    fetch_commit.id(),
                    true,
                    &format!("Setting {} to {}", remote_branch, fetch_commit.id()),
                )?;
                repo.set_head(&refname)?;
                repo.checkout_head(Some(
                    git2::build::CheckoutBuilder::default()
                        .allow_conflicts(true)
                        .conflict_style_merge(true)
                        .force(),
                ))?;
            }
        };
    } else if analysis.0.is_normal() {
        // do a normal merge
        let head_commit = repo.reference_to_annotated_commit(&repo.head()?)?;
        normal_merge(&repo, &head_commit, &fetch_commit)?;
    } else {
        info!("nothing to do...");
    }
    Ok(())
}

fn fast_forward(
    repo: &Repository,
    lb: &mut git2::Reference,
    rc: &git2::AnnotatedCommit,
) -> Result<(), git2::Error> {
    let name = match lb.name() {
        Some(s) => s.to_string(),
        None => String::from_utf8_lossy(lb.name_bytes()).to_string(),
    };
    let msg = format!("fast-forward: setting {} to id: {}", name, rc.id());
    info!("{}", msg);
    lb.set_target(rc.id(), &msg)?;
    repo.set_head(&name)?;
    repo.checkout_head(Some(
        git2::build::CheckoutBuilder::default()
            // For some reason the force is required to make the working directory actually get updated
            // I suspect we should be adding some logic to handle dirty working directory states
            // but this is just an example so maybe not.
            .force(),
    ))?;
    Ok(())
}

fn normal_merge(
    repo: &Repository,
    local: &git2::AnnotatedCommit,
    remote: &git2::AnnotatedCommit,
) -> Result<(), git2::Error> {
    let local_tree = repo.find_commit(local.id())?.tree()?;
    let remote_tree = repo.find_commit(remote.id())?.tree()?;
    let ancestor = repo
        .find_commit(repo.merge_base(local.id(), remote.id())?)?
        .tree()?;
    let mut idx = repo.merge_trees(&ancestor, &local_tree, &remote_tree, None)?;

    if idx.has_conflicts() {
        warn!("merge conficts detected...");
        repo.checkout_index(Some(&mut idx), None)?;
        return Ok(());
    }
    let result_tree = repo.find_tree(idx.write_tree_to(repo)?)?;
    // now create the merge commit
    let msg = format!("Merge: {} into {}", remote.id(), local.id());
    let sig = repo.signature()?;
    let local_commit = repo.find_commit(local.id())?;
    let remote_commit = repo.find_commit(remote.id())?;
    // Do our merge commit and set current branch head to that commit.
    let _merge_commit = repo.commit(
        Some("HEAD"),
        &sig,
        &sig,
        &msg,
        &result_tree,
        &[&local_commit, &remote_commit],
    )?;
    // Set working tree to match head.
    repo.checkout_head(None)?;
    Ok(())
}
