use std::{env::{current_dir, set_current_dir}, fs::{create_dir_all, read_to_string}, path::PathBuf};

use log::{debug, info, trace};
use serde::{Deserialize, Serialize};

use crate::{crates::Crate, prelude::abspath};

pub mod git;

#[derive(Serialize, Deserialize, Debug)]
#[serde(default)]
pub struct Source {
    interval: u64,
    git: Option<git::Git>,
}

impl Default for Source {
    fn default() -> Self {
        Source {
            interval: 60,
            git: None,
        }
    }
}

impl Source {
    pub fn available(&self) -> bool {
        if let Some(git) = &self.git {
            trace!("checking git...");
            return git.ensure().is_ok();
        }
        false
    }

    pub fn go_to_dir(&self) -> Result<(), Box<dyn std::error::Error>> {
        if let Some(git) = &self.git {
            if PathBuf::from(git.repository_path_str()) == current_dir()? {
                return Ok(())
            }
            let dir = git.ensure()?.repository_path()?;
            trace!("setting git dir as cwd... ({}@{}, {})", git.url, git.branch, dir.display());
            set_current_dir(dir)?;
        }
        Ok(())
    }

    pub fn get_crates(&self) -> Result<Vec<Crate>, Box<dyn std::error::Error>> {
        let mut crates = vec![];
        if let Some(git) = &self.git {
            trace!("getting crates from git...");
            debug!("{}", current_dir()?.display());

            // get crates (read dir, crates/*/crate.toml)
            for crate_file in glob::glob("crates/*/crate.toml").expect("err") {
                debug!("{crate_file:#?}");
                match crate_file {
                    Ok(cd) =>{ 
                        debug!("{}", cd.display());
                        crates.push(toml::from_str(&read_to_string(cd)?)?)
                    },
                    _ => continue
                }
            }
        }
        debug!("{:#?}", crates);
        Ok(crates)
    }
}
